# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import logging
import time
from random import randint

import zmq

from ..config import parse_config
from .common import CommonTransport

log = logging.getLogger(__name__)


def _encoded(func):
    def encoded(self, input):
        decoded = input.decode('utf-8')
        deserialized = json.loads(decoded)

        retval = func(self, deserialized)

        serialized = json.dumps(retval)
        out = serialized.encode('utf-8')
        return out
    return encoded

HEARTBEAT_LIVENESS = 3
HEARTBEAT_INTERVAL = 1
INTERVAL_INIT = 1
INTERVAL_MAX = 32

PPP_READY = b'\x01'
PPP_HEARTBEAT = b'\x02'
PPP_SHUTDOWN = b'\x03'


def worker_socket(addr, context, poller):
    worker = context.socket(zmq.DEALER)
    identity = "%04X-%04X" % (randint(0, 0x10000), randint(0, 0x10000))
    worker.setsockopt_string(zmq.IDENTITY, identity)
    poller.register(worker, zmq.POLLIN)
    worker.connect(addr)
    worker.send(PPP_READY)
    return worker


class ZmqTransport(CommonTransport):
    """ZmqTransport implements the Paranoid-Pirate worker described in the zguide:
    http://zguide.zeromq.org/py:all#Robust-Reliable-Queuing-Paranoid-Pirate-Pattern
    """

    def __init__(self, addr, context=None, registry=None):
        super().__init__(registry)
        if context is None:
            context = zmq.Context()

        self._addr = addr
        self._context = context

    def run(self):
        try:
            self._run()
        except KeyboardInterrupt:
            if self._worker:
                self._worker.send(PPP_SHUTDOWN)

    # the majority of this function is taken from:
    # http://zguide.zeromq.org/py:ppworker
    def _run(self):
        context = self._context
        poller = zmq.Poller()

        liveness = HEARTBEAT_LIVENESS
        interval = INTERVAL_INIT

        heartbeat_at = time.time() + HEARTBEAT_INTERVAL

        worker = worker_socket(self._addr, context, poller)
        self._worker = worker
        while True:
            socks = dict(poller.poll(HEARTBEAT_INTERVAL * 1000))

            # Handle worker activity on backend
            if socks.get(worker) == zmq.POLLIN:
                #  Get message
                #  - 3-part envelope + content -> request
                #  - 1-part HEARTBEAT -> heartbeat
                frames = worker.recv_multipart()
                if not frames:
                    break

                if len(frames) == 3:
                    client, empty, message = frames
                    assert empty == b''

                    log.debug('Recv message')
                    log.debug(message)
                    response = self.handle_message(message)
                    worker.send_multipart([
                        client,
                        b'',
                        response,
                    ])

                    liveness = HEARTBEAT_LIVENESS
                elif len(frames) == 1 and frames[0] == PPP_HEARTBEAT:
                    liveness = HEARTBEAT_LIVENESS
                else:
                    log.warn('Invalid message: %s', frames)
                interval = INTERVAL_INIT
            else:
                liveness -= 1
                if liveness == 0:
                    log.warn('Heartbeat failure, can\'t reach queue')
                    log.warn('Reconnecting in %0.2fs...' % interval)
                    time.sleep(interval)

                    if interval < INTERVAL_MAX:
                        interval *= 2
                    poller.unregister(worker)
                    worker.setsockopt(zmq.LINGER, 0)
                    worker.close()
                    worker = worker_socket(self._addr, context, poller)
                    self._worker = worker
                    liveness = HEARTBEAT_LIVENESS
            if time.time() > heartbeat_at:
                heartbeat_at = time.time() + HEARTBEAT_INTERVAL
                worker.send(PPP_HEARTBEAT)

    @_encoded
    def handle_message(self, req):
        kind = req['kind']
        param = req.get('param')
        if kind == 'init':
            parse_config(param.get('config') or {})
            return self.init_info()

        name = req['name']

        ctx = req.get('context') or {}

        if kind == 'provider':
            action = param.pop('action')
            return self.call_provider(ctx, name, action, param)
        elif kind == 'handler':
            return self.call_handler(ctx, name, param)
        else:
            return self.call_func(ctx, kind, name, param)



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import unittest

from ..error import SkygearException


class TestSkygearException(unittest.TestCase):
    def test_from_dict(self):
        e = SkygearException.from_dict({
            'message': 'This is an error.',
            'code': 24601,
            'info': {'reason': 'unknown'},
            })
        assert e.message == 'This is an error.'
        assert e.code == 24601
        assert e.info == {'reason': 'unknown'}

    def test_as_dict(self):
        e = SkygearException('This is an error.', 24601, {'reason': 'unknown'})
        d = e.as_dict()
        assert d == {
            'message': 'This is an error.',
            'code': 24601,
            'info': {'reason': 'unknown'},
            }



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import base64
import json
import logging
import os
from functools import wraps

from werkzeug.test import EnvironBuilder
from werkzeug.wrappers import BaseResponse, Request

from ..error import SkygearException
from ..registry import get_registry
from ..utils import db
from ..utils.context import start_context
from .encoding import _serialize_exc, deserialize_or_none, serialize_record


log = logging.getLogger(__name__)


def _get_engine():
    return db._get_engine()


def _wrap_result(f):
    @wraps(f)
    def wrapper(self, *args, **kwargs):
        try:
            return dict(result=f(self, *args, **kwargs))
        except Exception as e:
            handler = get_registry().get_exception_handler(e.__class__)
            if not handler:
                handler = handle_exception
            result = handler(e)
            if result is None:
                return dict(error=_serialize_exc(e).as_dict())
            elif isinstance(result, Exception):
                return dict(error=_serialize_exc(result).as_dict())
            else:
                return result
    return wrapper


def handle_exception(exc):
    if not isinstance(exc, SkygearException):
        log.exception("Error occurred processing request")
    return exc
get_registry().register_exception_handler(Exception, handle_exception)


def encode_base64_json(data):
    """
    Encode dict-like data into a base64 encoded JSON string.

    This can be used to get dict-like data into HTTP headers / envvar.
    """
    return base64.b64encode(bytes(json.dumps(data), 'utf-8'))


def decode_base64_json(data):
    """
    Decode dict-like data from a base64 encoded JSON string.

    This can be used to get dict-like data into HTTP headers / envvar.
    """
    return json.loads(base64.b64decode(data).decode('utf-8'))


def dict_from_base64_environ(name):
    data = os.environ.get(name)
    return decode_base64_json(data) if data else {}


class CommonTransport:
    def __init__(self, registry=None):
        self._registry = registry or get_registry()

    def init_info(self):
        return self._registry.func_list()

    @_wrap_result
    def call_func(self, ctx, kind, name, param):
        obj = self._registry.get_obj(kind, name)
        with start_context(ctx):
            if kind == 'op':
                return self.op(obj, param.get('args', {}))
            elif kind == 'hook':
                return self.hook(obj, param)
            elif kind == 'timer':
                return self.timer(obj)
            else:
                raise SkygearException("unknown plugin extension point")

    @_wrap_result
    def call_provider(self, ctx, name, action, param):
        obj = self._registry.get_obj('provider', name)

        with start_context(ctx):
            return self.provider(obj, action, param)

    @_wrap_result
    def call_handler(self, ctx, name, param):
        func = self._registry.get_handler(name, param['method'])
        with start_context(ctx):
            return self.handler(func, param)

    def handler(self, func, param):
        builder = EnvironBuilder(
            method=param['method'],
            path=param['path'],
            query_string=param.get('query_string'),
            headers=param['header'],
            data=base64.b64decode(param['body'])
        )
        environ = builder.get_environ()
        request = Request(environ, populate_request=False, shallow=False)
        response = func(request)
        status = 200
        if isinstance(response, BaseResponse):
            headers = {}
            for k, v in response.headers:
                headers[k] = [v]
            body_byte = response.get_data()
            body = base64.b64encode(body_byte).decode('utf-8')
            status = response.status_code
        elif isinstance(response, str):
            headers = {'Content-Type': ['text/plain; charset=utf-8']}
            body = base64.b64encode(
                bytes(response, 'utf-8')
            ).decode('utf-8')
        else:
            headers = {
                'Content-Type': ['application/json']
             }
            body = encode_base64_json(response).decode('utf-8')
        return {
            'status': status,
            'header': headers,
            'body': body
        }

    def op(self, func, param):
        if isinstance(param, list):
            args = param
            kwargs = {}
        elif isinstance(param, dict):
            args = []
            kwargs = param
        else:
            msg = "Unsupported args type '{0}'".format(type(param))
            raise ValueError(msg)
        return func(*args, **kwargs)

    def hook(self, func, param):
        original_record = deserialize_or_none(param.get('original', None))
        record = deserialize_or_none(param.get('record', None))
        with db.conn() as conn:
            returned = func(record, original_record, conn)

            # If the hook function does not return a value, assume that
            # the record in the first argument is to be returned.
            if returned is None:
                returned = record
        return serialize_record(returned)

    def timer(self, func):
        return func()

    def provider(self, provider, action, data):
        return provider.handle_action(action, data)

    def run(self):
        raise NotImplemented()



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from datetime import datetime

from ..models import Record, RecordID


class TestRecord():
    def test_create(eslf):
        now = datetime.now()
        rid = RecordID("note", "hello_world")
        r = Record(rid, "OWNER_ID", None,
                   created_at=now, created_by="CREATOR_ID",
                   updated_at=now, updated_by="UPDATER_ID",
                   data={"content": "Hello!", "noteOrder": 1})

        assert r is not None
        assert r.id == rid
        assert r.owner_id == 'OWNER_ID'
        assert r.created_at == now
        assert r.created_by == 'CREATOR_ID'
        assert r.updated_at == now
        assert r.updated_by == 'UPDATER_ID'
        assert r.data == {"content": "Hello!", "noteOrder": 1}

    def test_create_empty_data(eslf):
        rid = RecordID("note", "hello_world")
        r = Record(rid, "OWNER_ID", None)

        assert r is not None
        assert r.id == rid
        assert r.owner_id == 'OWNER_ID'
        assert r.data == {}



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from skygear.models import PublicAccessControlEntry, RoleAccessControlEntry

from ..encoding import deserialize_record, serialize_record


class TestsDeserializeRecord():
    def test_normal(self):
        rdata = {
            "_access": None,
            "_id": "note/99D92DBA-74D5-477F-B35E-F735E21B2DD5",
            "_ownerID": "OWNER_ID",
            "content": "Hello World!",
            "noteOrder": 1,
            }
        r = deserialize_record(rdata)
        assert r.id.type == "note"
        assert r.id.key == "99D92DBA-74D5-477F-B35E-F735E21B2DD5"
        print(r.data)
        assert r['content'] == "Hello World!"
        assert r['noteOrder'] == 1
        assert r.owner_id == "OWNER_ID"

    def test_model_with_acl(self):
        rdata = {
            "_id": "note/99D92DBA-74D5-477F-B35E-F735E21B2DD5",
            "_ownerID": "OWNER _ID",
            "_access": [{
                "role": "admin",
                "level": "write"
            }]
        }
        r = deserialize_record(rdata)
        assert r.id.type == "note"
        assert r.id.key == "99D92DBA-74D5-477F-B35E-F735E21B2DD5"
        assert isinstance(r.acl, list)
        assert isinstance(r.acl[0], RoleAccessControlEntry)

    def test_model_with_public_read_acl(self):
        rdata = {
            "_id": "note/99D92DBA-74D5-477F-B35E-F735E21B2DD5",
            "_ownerID": "OWNER _ID",
            "_access": [{
                "public": True,
                "level": "read"
            }]
        }
        r = deserialize_record(rdata)
        assert r.id.type == "note"
        assert r.id.key == "99D92DBA-74D5-477F-B35E-F735E21B2DD5"
        assert isinstance(r.acl, list)
        assert isinstance(r.acl[0], PublicAccessControlEntry)
        assert r.acl[0].level == "read"

    def test_serialize_record_back(self):
        rdata = {
            "_id": "note/99D92DBA-74D5-477F-B35E-F735E21B2DD5",
            "_ownerID": "OWNER_ID",
            "_access": [{
                "public": True,
                "level": "read"
            }],
            "_created_at": "2014-09-27T17:40:00Z"
        }
        r = deserialize_record(rdata)
        result = serialize_record(r)
        assert result['_id'] == "note/99D92DBA-74D5-477F-B35E-F735E21B2DD5"
        assert result['_ownerID'] == "OWNER_ID"
        assert result['_access'] == [{
            "public": True,
            "level": "read"
        }]
        assert result['_created_at'] == "2014-09-27T17:40:00Z"



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import logging

from werkzeug.routing import BaseConverter, Map, Rule
from werkzeug.serving import run_simple
from werkzeug.wrappers import Request, Response

from ..config import parse_config
from .common import CommonTransport, decode_base64_json
from .encoding import _serialize_exc


log = logging.getLogger(__name__)


class RegexConverter(BaseConverter):
    """
    This converter matches URL by regular expression.

    Copied from http://stackoverflow.com/questions/5870188/does-flask-support-regular-expressions-in-its-url-routing  #noqa
    """
    def __init__(self, url_map, *items):
        super(RegexConverter, self).__init__(url_map)
        self.regex = items[0]


class HttpTransport(CommonTransport):
    """
    HttpTransport implements a transport protocol between skygear and
    plugin that communicates over a HTTP connection.
    """

    def _url_map(self):
        """
        Returns a URL routing map.
        """
        return Map([
            Rule('/init', endpoint='init'),
            Rule('/op/<name>', endpoint='op'),
            Rule('/handler/<regex(".*"):name>', endpoint='handler'),
            Rule('/hook/<name>', endpoint='hook'),
            Rule('/provider/<name>/<action>', endpoint='provider'),
            Rule('/timer/<name>', endpoint='timer'),
            ], converters={'regex': RegexConverter})

    def __init__(self, addr, registry=None, debug=False):
        super().__init__(registry)
        self.url_map = self._url_map()
        if ':' in addr:
            hostname, port = addr.split(':', 2)
        else:
            hostname = addr
            port = 8000
        self.hostname = hostname.strip() if hostname.strip() else '0.0.0.0'
        try:
            self.port = int(port)
        except:
            self.port = 8000
        self.debug = debug

    @Request.application
    def dispatch(self, request):
        """
        Application from the view of the plugin web server.
        """

        # Returns an HTTP response in JSON format
        try:
            output = self._dispatch(request)
        except Exception as e:
            log.exception("exception while handling request")
            output = dict(error=_serialize_exc(e).as_dict())
        return Response(json.dumps(output), mimetype="application/json")

    def _dispatch(self, request):
        """
        Dispatches request to a plugin extension point function.
        """
        endpoint, values, context, param = self.read_request(request)
        if endpoint == 'init':
            parse_config(param.get('config', {}))
            return self.init_info()
        else:
            if endpoint == 'provider':
                return self.call_provider(context, values['name'],
                                          values['action'], param)
            elif endpoint == 'handler':
                return self.call_handler(context, values['name'], param)
            else:
                return self.call_func(context, endpoint, values['name'], param)

    def read_request(self, request):
        """
        Reads and returns HTTP request.
        """
        adapter = self.url_map.bind_to_environ(request.environ)
        endpoint, values = adapter.match()

        request_data = request.get_data(as_text=True)
        param = json.loads(request_data) if request_data else {}

        context_data = request.headers.get('X-Skygear-Plugin-Context')
        context = decode_base64_json(context_data) if context_data else {}

        return endpoint, values, context, param

    def run(self):
        """
        Start the web server.
        """
        run_simple(self.hostname, self.port, self.dispatch,
                   threaded=True,
                   use_reloader=self.debug)



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import configargparse as argparse

options = argparse.Namespace()


def get_argument_parser():
    ap = argparse.ArgumentParser(description='Skygear Plugin runner')
    ap.add_argument('--skygear-address', metavar='ADDR', action='store',
                    default='tcp://127.0.0.1:5555',
                    help="Binds to this socket for skygear",
                    env_var='SKYGEAR_ADDRESS')
    ap.add_argument('--skygear-endpoint', metavar='ENDPOINT', action='store',
                    default='http://127.0.0.1:3000',
                    help="Send to this addres for skygear handlers",
                    env_var='SKYGEAR_ENDPOINT')
    ap.add_argument('--apikey', metavar='APIKEY', action='store',
                    default=None,
                    help="API Key of the application",
                    env_var='SKYGEAR_APIKEY')
    ap.add_argument('--masterkey', metavar='MASTERKEY', action='store',
                    default=None,
                    help="Master Key of the application",
                    env_var='SKYGEAR_MASTERKEY')
    ap.add_argument('--appname', metavar='APPNAME', action='store',
                    default='',
                    help="Application name of the skygear daemon",
                    env_var='SKYGEAR_APPNAME')
    ap.add_argument('--loglevel', action='store', default='INFO',
                    help="Log level",
                    env_var='SKYGEAR_LOGLEVEL')
    ap.add_argument('--subprocess', dest='subprocess', action='store',
                    nargs='+',
                    metavar=('(init|op|hook|handler|timer)', 'name'),
                    help='Trigger subprocess everytime for debug')
    ap.add_argument('--http', action='store_true',
                    help='Trigger http web server',
                    env_var='SKYGEAR_HTTP')
    ap.add_argument('--http-addr', metavar='HTTP_ADDR', action='store',
                    default='0.0.0.0:8000',
                    help='Address where htp web server listen to',
                    env_var='SKYGEAR_HTTP_ADDR')
    ap.add_argument('--debug', action='store_true',
                    help='Enable debugging features',
                    env_var='SKYGEAR_DEBUG')
    ap.add_argument('--collect-assets', metavar='DIST', action='store',
                    help="Collect static assets to a directory")
    ap.add_argument('--force-assets', action='store_true',
                    help="Remove dist folder before proceeding")
    ap.add_argument('--serve-static-assets', action='store_true',
                    env_var='SKYGEAR_SERVE_STATIC_ASSETS',
                    help="Enable to serve static asset from plugin process")
    ap.add_argument('plugin', nargs='?')
    return ap


def parse_args():
    global options
    options = get_argument_parser().parse_args(namespace=options)
    return options



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import datetime
import unittest
from unittest.mock import patch

from ..container import SkygearContainer, send_action


class TestSendAction(unittest.TestCase):
    @patch('skygear.container.requests', autospec=True)
    def test_send_str(self, mock_requests):
        send_action('http://skygear.dev/', {
            'key': 'string'
        })
        self.assertEqual(len(mock_requests.method_calls), 1)
        call = mock_requests.method_calls[0]
        self.assertEqual(call[0], 'post')
        self.assertEqual(call[1][0], 'http://skygear.dev/')
        self.assertEqual(call[2]['data'], '{"key": "string"}')

    @patch('skygear.container.requests', autospec=True)
    def test_send_date(self, mock_requests):
        dt = datetime.datetime(2014, 9, 27, 17, 40, 0,
                               tzinfo=datetime.timezone.utc)
        send_action('http://skygear.dev/', {
            'print_at': dt
        })
        self.assertEqual(len(mock_requests.method_calls), 1)
        call = mock_requests.method_calls[0]
        self.assertEqual(call[0], 'post')
        self.assertEqual(call[1][0], 'http://skygear.dev/')
        self.assertEqual(
            call[2]['data'],
            '{"print_at": "2014-09-27T17:40:00Z"}')


class TestContainer():
    def test_payload_include_action(self):
        c = SkygearContainer(endpoint='endpoint', access_token='access-token')
        payload = c._payload('action:work', {'data': 'hello world'})
        assert payload['action'] == 'action:work'
        assert payload['data'] == 'hello world'

    def test_payload_include_api_key(self):
        c = SkygearContainer(endpoint='endpoint', api_key='api-key')
        payload = c._payload('action:work', {})
        assert payload['api_key'] == 'api-key'

    def test_payload_include_user_id(self):
        c = SkygearContainer(endpoint='endpoint', api_key='api-key',
                             user_id='user-id')
        payload = c._payload('action:work', {})
        assert payload['_user_id'] == 'user-id'

    def test_payload_include_access_token(self):
        c = SkygearContainer(endpoint='endpoint', access_token='access-token')
        payload = c._payload('action:work', {})
        assert payload['access_token'] == 'access-token'

    def test_payload_include_all_credentials(self):
        c = SkygearContainer(endpoint='endpoint',
                             access_token='access-token',
                             api_key='api-key',
                             user_id='user-id')
        payload = c._payload('action:work', {})
        assert payload['access_token'] == 'access-token'
        assert payload['api_key'] == 'api-key'
        assert payload['_user_id'] == 'user-id'



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from unittest.mock import patch

from .. import config


class TestParseConfig():
    def test_simple_dict(self):
        ns = config.Configuration()
        assert config._parse_config({'hello': 'world'}, ns) == ns
        assert ns.hello == 'world'

    def test_complex_dict(self):
        ns = config.Configuration()
        data = {
                'hello': 'world',
                'bye': {
                    'apple': {'fruit': True},
                    'pie': {'fruit': False},
                    }
                }
        assert config._parse_config(data, ns) == ns
        assert ns.hello == 'world'
        assert isinstance(ns.bye, config.Configuration)
        assert isinstance(ns.bye.apple, config.Configuration)
        assert isinstance(ns.bye.pie, config.Configuration)
        assert ns.bye.apple.fruit is True
        assert ns.bye.pie.fruit is False

    @patch('skygear.config.config', config.Configuration())
    def test_parse_config(self):
        assert config.parse_config({'hello': 'world'}) == config.config
        assert config.config.hello == 'world'



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import logging
import os
from urllib.parse import urlparse, urlunparse

from websocket import create_connection

from .options import options

encoder = json.dumps
_hub = None


def get_hub():
    global _hub
    if not _hub:
        _hub = Hub()
    return _hub


def publish(channel, data):
    get_hub().publish(channel, data)


def _get_default_pubsub_url():
    if os.getenv('PUBSUB_URL', None):
        logger = logging.getLogger(__name__)
        logger.info('Environment variable PUBSUB_URL ("{0}") is deprecated, '
                    'but this setting is honored anyway. To remove this '
                    'message, remove PUBSUB_URL from environment and make '
                    'sure you specify skygear endpoint and API key.'
                    .format(os.getenv('PUBSUB_URL')))
        return os.getenv('PUBSUB_URL')

    parsed_endpoint = urlparse(options.skygear_endpoint)

    scheme = 'wss' if parsed_endpoint.scheme == 'https' else 'ws'
    netloc = parsed_endpoint.netloc
    path = '/pubsub'
    urlparts = (scheme, netloc, path, '', '', '')
    return urlunparse(urlparts)


class Hub:

    def __init__(self, end_point=None, api_key=None):
        self.transport = 'websocket'
        self.end_point = end_point or _get_default_pubsub_url()
        self.api_key = api_key or options.apikey

    def publish(self, channel, data):
        wsopts = {}
        if self.api_key:
            wsopts['header'] = [
                    'X-Skygear-API-Key: {0}'.format(self.api_key)
                    ]
        conn = create_connection(self.end_point, **wsopts)
        _data = encoder({
            'action': 'pub',
            'channel': channel,
            'data': data,
        })
        conn.send(_data)
        conn.close()



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import datetime
from functools import wraps

from .registry import get_registry
from .utils.assets import DirectoryStaticAssetsLoader, StaticAssetsLoader

_registry = get_registry()


def op(name, *args, **kwargs):
    def our_op(func):
        _registry.register("op", name, func, *args, **kwargs)
        return func
    return our_op


def every(interval, name=None, *args, **kwargs):
    if isinstance(interval, datetime.timedelta):
        interval = "@every {0}s".format(interval.total_seconds())
    elif isinstance(interval, int):
        interval = "@every {0}s".format(interval)
    elif not isinstance(interval, str):
        msg = "Expecting int, timedelta or str for interval. Got '{0}'." \
            .format(type(interval).__name__)
        raise Exception(msg)
    kwargs['spec'] = interval
    kwargs['name'] = name

    def our_every(func):
        name = kwargs.pop('name', None) or \
            func.__module__ + "." + func.__name__

        _registry.register("timer", name, func, *args, **kwargs)
        return func
    return our_every


def handler(name, *args, **kwargs):
    def skygear_handler(func):
        _registry.register("handler", name, func, *args, **kwargs)
        return func
    return skygear_handler


def hook(trigger, *args, **kwargs):
    kwargs['trigger'] = trigger

    def skygear_hook(func):
        def hook_func(record, original_record, db):
            # return the record for user
            func(record, original_record, db)
            return record

        name = kwargs.pop('name', None) or \
            func.__module__ + "." + func.__name__

        _registry.register("hook", name, hook_func, *args, **kwargs)
        return func

    return skygear_hook


def register_save_hook(func, *args, **kwargs):
    def hook_func(record, original_record, db):
        func(record, original_record, db)
        return record

    name = kwargs.pop('name', None) or \
        func.__module__ + "." + func.__name__

    _registry.register("hook", name, hook_func, *args, **kwargs)


def register_delete_hook(func, *args, **kwargs):
    def hook_func(record, original_record, db):
        func(record, db)
        return record

    name = kwargs.pop('name', None) or \
        func.__module__ + "." + func.__name__

    _registry.register("hook", name, hook_func, *args, **kwargs)


def before_save(type_, *args, **kwargs):
    kwargs['type'] = type_
    kwargs['trigger'] = "beforeSave"

    def wrapper(func):
        register_save_hook(func, *args, **kwargs)
        return func
    return wrapper


def after_save(type_, *args, **kwargs):
    kwargs['type'] = type_
    kwargs['trigger'] = "afterSave"

    def wrapper(func):
        register_save_hook(func, *args, **kwargs)
        return func
    return wrapper


def before_delete(type_, *args, **kwargs):
    kwargs['type'] = type_
    kwargs['trigger'] = "beforeDelete"

    def wrapper(func):
        register_delete_hook(func, *args, **kwargs)
        return func
    return wrapper


def after_delete(type_, *args, **kwargs):
    kwargs['type'] = type_
    kwargs['trigger'] = "afterDelete"

    def wrapper(func):
        register_delete_hook(func, *args, **kwargs)
        return func
    return wrapper


def provides(provider_type, provider_id, *args, **kwargs):
    def skygear_provider(klass):
        provider = klass()
        _registry.register_provider(provider_type, provider_id, provider,
                                    *args, **kwargs)
        return klass
    return skygear_provider


def _fix_handler_path(name):
    name = name.replace(':', '/')
    if name.startswith('/'):
        name = name[1:]
    if name.endswith('/'):
        name = name[:-1]
    return name


def rest(name, *args, **kwargs):
    name = _fix_handler_path(name)

    def wrapper(klass):
        if callable(getattr(klass, 'get_allowed_methods', None)):
            kwargs['method'] = klass.get_allowed_methods()
        else:
            kwargs['method'] = ['GET', 'POST', 'PUT', 'DELETE']

        base_name = name if name.startswith('/') else '/' + name

        @handler(name, *args, **kwargs)
        @handler(name + '/', *args, **kwargs)
        def call_restful_wraps(request):
            restful = klass()
            return restful.handle_request(base_name, request)
        return klass
    return wrapper


def static_assets(prefix, *args, **kwargs):
    """
    This decorator is used to decorate a function that returns
    information of static assets.

    The decorated function should return the absolute path of the static
    assets directory.
    """
    def wrapper(func):
        @wraps(func)
        def wrap_loader():
            loader = func()
            if not loader:
                return None
            if not isinstance(loader, StaticAssetsLoader):
                loader = DirectoryStaticAssetsLoader(loader)
            return loader
        _registry.register_static_assets(prefix, wrap_loader)
        return wrap_loader
    return wrapper


def exception_handler(klass):
    """
    This decorator is used to decorate a function that is called by when
    exception with the specified class has occurred.

    The decorated function must have one parameter, which is the exception
    that occurred. The decorated function is expected to return an exception
    or a result. If the decorated function does not return a value, the
    original exception is returned instead.

    When an exception handler for a particular exception class already exists,
    the newly decorated function for the same exception class will override
    previously registered exception handler. If there are multiple exception
    handlers, the handler with the most specific exception class will
    be selected to handle an exception.
    """
    def wrapper(func):
        def wrap(*args, **kwargs):
            # not called
            pass

        _registry.register_exception_handler(klass, func)
        return func
    return wrapper



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import inspect
import os.path
import shutil


def _trim_abs_path(path):
    if path.startswith('/'):
        path = path[1:]
    if path.startswith('./'):
        path = path[2:]
    return path


class StaticAssetsLoader:
    """
    This loader is a generic class for loading static assets.
    """
    def get_asset(self, name):
        """
        Get content of a static assets by name.
        """
        return None

    def copy_into(self, path):
        """
        Copy the content of all static assets into a directory.
        """
        pass

    def exists_asset(self, name):
        return False


class DictStaticAssetsLoader(StaticAssetsLoader):
    """
    This class loads static asset from a provided dictionary.
    """
    def __init__(self, _dict):
        super().__init__()
        self._dict = _dict

    def get_asset(self, name):
        return self._dict.get(name, None)

    def copy_into(self, path):
        for filename, content in self._dict.items():
            filepath = os.path.abspath(
                    os.path.join(path, _trim_abs_path(filename)))
            os.makedirs(os.path.dirname(filepath), exist_ok=True)
            with open(filepath, 'wb') as f:
                f.write(content)

    def exists_asset(self, name):
        return name in self._dict


class DirectoryStaticAssetsLoader(StaticAssetsLoader):
    """
    This class loads static asset from a provided file system directory.
    """
    def __init__(self, dirpath):
        super().__init__()
        self._dirpath = os.path.abspath(dirpath)

    @property
    def dirpath(self):
        return self._dirpath

    def get_asset(self, path):
        filename = os.path.join(self.dirpath, path)
        with open(filename, 'rb') as f:
            return f.read()

    def copy_into(self, dest):
        shutil.copytree(self.dirpath, dest, symlinks=True)

    def exists_asset(self, name):
        return os.path.exists(os.path.join(self.dirpath, name))


class PackageStaticAssetsLoader(StaticAssetsLoader):
    """
    This class loads static asset from the provided python package.
    """
    def __init__(self, package_name, package_path):
        super().__init__()
        from pkg_resources import ResourceManager, get_provider
        self.provider = get_provider(package_name)
        self.manager = ResourceManager()
        self.package_path = package_path

    def resource_name(self, name):
        return os.path.join(self.package_path, _trim_abs_path(name))

    def get_asset(self, name):
        return self.provider.get_resource_string(self.manager,
                                                 self.resource_name(name))

    def copy_into(self, dest):
        def _walk(subpath=None):
            """
            `_walk` copies files in a subdirectory to the destination. If
            file is a directory, recursively call `_walk` for the directory.

            If subpath evaluates to False, copy from the `package_path`.
            """
            if subpath:
                resource_dir = self.resource_name(subpath)
            else:
                resource_dir = self.package_path

            for filename in self.provider.resource_listdir(resource_dir):
                childpath = subpath + '/' + filename if subpath else filename
                dest_name = os.path.abspath(os.path.join(dest, childpath))
                if self.provider.resource_isdir(self.resource_name(childpath)):
                    os.mkdir(dest_name)
                    _walk(childpath)
                else:
                    print('writing to {}'.format(dest_name))
                    with open(dest_name, 'wb') as f:
                        f.write(self.get_asset(childpath))

        os.makedirs(dest, exist_ok=True)
        _walk()

    def exists_asset(self, name):
        return self.provider.has_resource(self.resource_name(name))


def directory_assets(path='static'):
    """
    This helper function returns the absolute path of the static assets
    directory by specifying a path relative to the current directory,
    or by specifying a absolute path.
    """
    return DirectoryStaticAssetsLoader(path)


def relative_assets(path='static', current_file=None):
    """
    This helper function returns the absolute path of the static assets
    directory by specifying a path relative to the file of a python module.

    If the file of a python module is not specified, the `__file__` of the
    caller module is used. This is effectively the same as calling
    this function with `relative_assets(path, __file__)`.
    """
    if not current_file:
        frame = inspect.stack()[1]
        module = inspect.getmodule(frame[0])
        current_file = module.__file__
    rel_dir = os.path.dirname(current_file)
    return directory_assets(os.path.join(rel_dir, path))


def package_assets(package_name, path='static'):
    """
    This helper function returns the absolute path of the static assets
    directory by specifying a path relative to a specified package name.
    """
    return PackageStaticAssetsLoader(package_name, path)



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
NotAuthenticated = 101
PermissionDenied = 102
AccessKeyNotAccepted = 103
AccessTokenNotAccepted = 104
InvalidCredentials = 105
InvalidSignature = 106
BadRequest = 107
InvalidArgument = 108
Duplicated = 109
ResourceNotFound = 110
NotSupported = 111
NotImplemented = 112
ConstraintViolated = 113
IncompatibleSchema = 114
AtomicOperationFailure = 115
PartialOperationFailure = 116
UndefinedOperation = 117
PluginUnavailable = 118
PluginTimeout = 119

UnexpectedError = 10000


class SkygearException(Exception):
    """
    An exception object to be raised by plugin to indicate that
    the requested operation is not successful, or that the operation
    is aborted for whatever reason.
    """
    def __init__(self, message, code=UnexpectedError, info=None):
        self.message = message
        self.code = code
        self.info = info or {}
        super(SkygearException, self).__init__(message)

    def as_dict(self):
        return {
            'message': self.message,
            'code': self.code,
            'info': self.info,
            }

    @classmethod
    def from_dict(cls, error_dict):
        message = error_dict.get('message', "An error has occurred.")
        code = error_dict.get('code', UnexpectedError)
        info = error_dict.get('info', {})
        return cls(message, code, info)



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os.path
import shutil
import tempfile
import unittest

from .. import assets as assetsutils


class StaticAssetsHelperFunction(unittest.TestCase):
    def test_directory_assets(self):
        loader = assetsutils.directory_assets('hello-world')
        assert isinstance(loader, assetsutils.StaticAssetsLoader)
        assert loader.dirpath == os.path.abspath('hello-world')

    def test_relative_assets(self):
        loader = assetsutils.relative_assets('hello-world')
        assert isinstance(loader, assetsutils.StaticAssetsLoader)
        expected = os.path.join(os.path.dirname(__file__), 'hello-world')
        assert loader.dirpath == os.path.abspath(expected)

    def test_package_assets(self):
        loader = assetsutils.package_assets(__name__, 'assets')
        assert isinstance(loader, assetsutils.PackageStaticAssetsLoader)


class TestPackageStaticAssetsLoader(unittest.TestCase):
    def setUp(self):
        self.loader = assetsutils.PackageStaticAssetsLoader(__name__, 'assets')
        self.dist = tempfile.mkdtemp()

    def tearDown(self):
        if os.path.exists(self.dist):
            shutil.rmtree(self.dist)

    def test_get_asset(self):
        content = self.loader.get_asset('content/index.txt')
        assert content.startswith(b'Hello World!')

    def test_copy_into(self):
        self.loader.copy_into(self.dist)
        with open(os.path.join(self.dist, 'content/index.txt'), 'rb') as f:
            assert f.read().startswith(b'Hello World!')

    def test_exists_asset(self):
        assert self.loader.exists_asset('content/index.txt') is True

    def test_exists_asset_nonexistent(self):
        assert self.loader.exists_asset('non-existent') is False



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import logging
import sys

from .. import config as skyconfig
from .. import error as skyerr
from .common import CommonTransport, dict_from_base64_environ

log = logging.getLogger(__name__)


ACCEPTED_TARGETS = ['init', 'op', 'hook', 'handler', 'timer', 'provider']


class ConsoleTransport(CommonTransport):
    def __init__(self, args, stdin=sys.stdin, stdout=sys.stdout,
                 registry=None):
        super().__init__(registry)
        self.args = args

        self.input = stdin
        self.output = stdout

    def run(self):
        self.args = self.args
        target = self.args[0]
        if target not in ACCEPTED_TARGETS:
            log.error("Only init, op, hook, handler, timer and provider is"
                      " supported now")
            sys.exit(1)
        if target == 'init':
            self.writeJSON(self.init_info())
        elif len(self.args) < 2:
            log.error("Missing param for %s", target)
            sys.exit(1)
        else:
            try:
                self.handle_command(target, self.args)
            except skyerr.SkygearException as e:
                self.writeJSON(e.as_dict())

    def handle_command(self, target, args):
        param = self.readJSON()
        context = dict_from_base64_environ('SKYGEAR_CONTEXT')
        skyconfig.parse_config(dict_from_base64_environ('SKYGEAR_CONFIG'))

        if target == 'provider':
            output = self.call_provider(context, args[1], args[2], param)
        elif target == 'handler':
            output = self.call_handler(context, args[1], param)
        else:
            output = self.call_func(context, target, args[1], param)

        self.writeJSON(output)

    def readJSON(self):
        data = self.read()
        if not data:
            return {}
        try:
            return json.loads(data)
        except ValueError:
            msg = "unable to parse JSON string"
            logging.exception(msg)
            raise skyerr.SkygearException(msg, skyerr.UnexpectedError)

    def writeJSON(self, data):
        try:
            self.write(json.dumps(data))
        except TypeError:
            msg = "unable to serialize obj to JSON string"
            logging.exception(msg)
            raise skyerr.SkygearException(msg, skyerr.UnexpectedError)

    def read(self):
        if not self.input.isatty():
            return "".join(self.input)
        else:
            return ""

    def write(self, obj, format='text'):
        return self.output.write(obj)



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class BaseAuthProvider(object):
    def handle_action(self, action, data):
        auth_data = data.get('auth_data', {})
        if action == 'login':
            output = self.login(auth_data)
        elif action == 'logout':
            output = self.logout(auth_data)
        elif action == 'info':
            output = self.info(auth_data)
        return output

    def login(self, auth_data):
        raise NotImplementedError(
            "Subclass of BaseAuthProvider should implement login method.")

    def logout(self, auth_data):
        raise NotImplementedError(
            "Subclass of BaseAuthProvider should implement logout method.")

    def info(self, auth_data):
        raise NotImplementedError(
            "Subclass of BaseAuthProvider should implement info method.")



import logging
import os

from ..assets import CollectorException, StaticAssetsCollector
from ..options import options
from ..registry import get_registry
from ..utils.assets import StaticAssetsLoader

_registry = get_registry()
log = logging.getLogger(__name__)


def collect_static_assets():
    dist = os.path.abspath(os.path.expanduser(options.collect_assets))
    collector = StaticAssetsCollector(dist)

    if os.path.exists(dist):
        if not options.force_assets:
            log.error('Directory %s already exists. Remove the directory '
                      'first, or specify --force-assets to discard files in '
                      'the directory.',
                      dist)
            exit(1)
        collector.clean()

    if not _registry.static_assets:
        log.warn('No static assets are declared.')
        return collector.dist

    for prefix, func in _registry.static_assets.items():
        loader = func()
        if not loader:
            continue
        elif not isinstance(loader, StaticAssetsLoader):
            raise ValueError('Function decorated with static_assets '
                             'must return a string or '
                             'an instance of StaticAssetsLoader')
        try:
            log.info('Static assets %s', prefix)
            collector.collect(prefix, loader)
        except CollectorException as e:
            log.error(str(e))
            exit(1)

    return collector.dist



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from .console import ConsoleTransport
from .http import HttpTransport

try:
    from .zmq import ZmqTransport
except ImportError:
    from .common import CommonTransport

    class ZmqTransport(CommonTransport):
        """A dummy ZmqTransport class to raise a proper exception
        """

        def __init__(self, addr, context=None, registry=None):
            raise ImportError(
                'zmq transport is not installed. '
                'Please install via `pip install --upgrade skygear[zmq]`')

__all__ = ['ConsoleTransport', 'HttpTransport', 'ZmqTransport']



# Copyright 2015 Oursky Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import argparse


class Configuration(argparse.Namespace):
    pass


config = Configuration()


def _parse_config(data, ns):
    for k, v in data.items():
        if isinstance(v, dict):
            setattr(ns, k, _parse_config(v, Configuration()))
        else:
            setattr(ns, k, v)
    return ns


def parse_config(data, namespace=None):
    """
    Parse the Skygear config from a dict to a Configuration object.

    When namespace is not specified, the global config is updated instead.
    """
    global config
    return _parse_config(data, namespace or config)
